#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <map>
#include <CAVERN.hxx>
#include "ygDebugFlags.h"
#include "ygMutex.h"
#include "ygUtil.h"
#include "ygNetClient.h"

using namespace std;


class ygNetClientKey
	{
	public:
	 void *data;
	 size_t size;
	 size_t bufferSize;
	 bool owned;
	};

typedef map<ygString,ygNetClientKey *> ygNetClientKeymap;
typedef map<ygString,ygNetClientKeymap> ygNetClientPathmap;


struct _ygNetClientPrivateData
	{
	ygString server;
	int port;
	CAVERNnet_udp_c * udpSocket;
	CAVERNnet_tcpReflectorClient_c * tcpSocket;
	ygNetClientCallback callback;
	void * callbackData;
	ygNetClientPathmap paths;
	CAVERNnet_datapack_c packer;
	char * tcpBuffer;
	void * packetBuffer;
	ygMutex mapMutex, dataMutex;
	ygNetClientPathmap::const_iterator refreshIter;
	float refreshInterval, refreshTime;
	bool debugDownload;
	};



ygNetClient::ygNetClient(void)
	{
	p_ = new struct _ygNetClientPrivateData;
	p_->udpSocket = NULL;
	p_->tcpSocket = NULL;
	p_->tcpBuffer = NULL;
	p_->callback = NULL;
	p_->packetBuffer = malloc(YG_NET_MAX_BUFFER_SIZE);
	p_->refreshIter = p_->paths.begin();
	if (getenv("YG_NET_REFRESH_INTERVAL"))
		p_->refreshInterval = atof(getenv("YG_NET_REFRESH_INTERVAL"));
	else
		p_->refreshInterval = -1;
	p_->refreshTime = 0;
	p_->debugDownload = ygDebugFlags::checkDebugEnv("net.download");
	}


void ygNetClient::init(const ygString& server,int port)
	{
	p_->server = server;
	p_->port = port;
	p_->udpSocket = new CAVERNnet_udp_c;
	p_->udpSocket->init(0);
	p_->udpSocket->setSendAddress(server.c_str(),port);
	p_->udpSocket->makeNonBlocking();
	p_->tcpSocket = new CAVERNnet_tcpReflectorClient_c;
	if (p_->tcpSocket->connectToServer(server.c_str(),port) ==
			CAVERNnet_tcpReflectorClient_c::FAILED)
		{
		cerr << "ERROR: ygNetClient::init: failed to connect"
			" to TCP server " << server << ":" << port << endl;
		return;
		}
	ygString cacheServer(server);
	int cachePort = port+1;
	if (getenv("YG_NET_CACHE_SERVER"))
		cacheServer = getenv("YG_NET_CACHE_SERVER");
	if (getenv("YG_NET_CACHE_PORT"))
		cachePort = atoi(getenv("YG_NET_CACHE_PORT"));
	downloadViaTCP(cacheServer,cachePort);
	p_->udpSocket->enableInstrumentation();
	}


void ygNetClient::showStats(void)
	{
	p_->udpSocket->showStats("UDP","ygNetClient UDP socket");
	p_->tcpSocket->showStats("TCP","ygNetClient TCP socket");
	}


void ygNetClient::reset(void)
	{
	if (p_->tcpSocket)
		p_->tcpSocket->close();
	else
		p_->tcpSocket = new CAVERNnet_tcpReflectorClient_c;
	if (p_->tcpSocket->connectToServer(p_->server.c_str(),p_->port) ==
			CAVERNnet_tcpReflectorClient_c::FAILED)
		{
		cerr << "ERROR: ygNetClient::reset: failed to connect"
			" to TCP server " << p_->server << ":" << p_->port << endl;
		return;
		}
	ygString cacheServer(p_->server);
	int cachePort = p_->port+1;
	if (getenv("YG_NET_CACHE_SERVER"))
		cacheServer = getenv("YG_NET_CACHE_SERVER");
	if (getenv("YG_NET_CACHE_PORT"))
		cachePort = atoi(getenv("YG_NET_CACHE_PORT"));
	downloadViaTCP(cacheServer,cachePort);
	uploadViaTCP();
	}


void ygNetClient::trigger(ygNetClientCallback cb,void *data)
	{
	p_->callback = cb;
	p_->callbackData = data;
	}


void * ygNetClient::get(const ygString& path,const ygString& key,size_t *size)
	{
	ygNetClientKey * keystruct = getKeyStruct(path,key,false);
	if ((!keystruct) || (keystruct->size <= 0))
		{
		*size = 0;
		return 0;
		}
	p_->dataMutex.lock();
	*size = keystruct->size;
	void * retData = malloc(keystruct->size);
	memcpy(retData, keystruct->data, keystruct->size);
	p_->dataMutex.unlock();
	return retData;
	}


ygNetClientKey * ygNetClient::getKeyStruct(const ygString& path, const ygString& key,bool owned)
	{
	p_->mapMutex.lock();
	ygNetClientKey * keystruct = p_->paths[path][key];
	if (!keystruct)
		{
		keystruct = new ygNetClientKey;
		keystruct->data = NULL;
		keystruct->size = 0;
		keystruct->bufferSize = 0;
		keystruct->owned = owned;
		p_->paths[path][key] = keystruct;
		}
	p_->mapMutex.unlock();
	return keystruct;
	}


void ygNetClient::put(const ygString& path, const ygString& key, void *data, size_t size)
	{
	saveKeyData(path,key,data,size);
	sendKeyDataTCP(path,key,data,size);
	}


void ygNetClient::putUnreliably(const ygString& path, const ygString& key, void *data, size_t size)
	{
	saveKeyData(path,key,data,size);
	sendKeyDataUDP(path,key,data,size);
	}


void ygNetClient::saveKeyData(const ygString& path, const ygString& key, void *data, size_t size)
	{
	ygNetClientKey * keystruct = getKeyStruct(path,key,true);
	if (keystruct)
		{
		p_->dataMutex.lock();
		if (size > keystruct->bufferSize)
			{
			keystruct->bufferSize = size;
			keystruct->data = realloc(keystruct->data,keystruct->bufferSize);
			}
		memcpy(keystruct->data,data,size);
		keystruct->size = size;
		p_->dataMutex.unlock();
		}
	}


void ygNetClient::sendKeyDataTCP(const ygString& path, const ygString& key, void *data, size_t size, int sendType)
	{
	if (p_->tcpSocket)
		{
		int totalSize = sizeof(int) + 
				sizeof(int) + path.length()+1 +
				sizeof(int) + key.length()+1 +
				sizeof(int) + size;
		initPacker(totalSize);
		p_->packer.packInt(sendType);
		p_->packer.packInt(path.length()+1);
		p_->packer.pack(path.c_str(), path.length()+1);
		p_->packer.packInt(key.length()+1);
		p_->packer.pack(key.c_str(), key.length()+1);
		p_->packer.packInt(size);
		p_->packer.pack((char *)data, size);
		int nbytes = p_->packer.getBufferFilledSize();
		p_->tcpSocket->write(p_->packer.getBuffer(),&nbytes);
		}
	else
		sendKeyDataUDP(path, key, data, size, sendType);
	}


void ygNetClient::sendKeyDataUDP(const ygString& path, const ygString& key, void *data, size_t size, int sendType)
	{
	if (p_->udpSocket)
		{
		int totalSize = sizeof(int) + 
				sizeof(int) + path.length()+1 +
				sizeof(int) + key.length()+1 +
				sizeof(int) + size;
		initPacker(totalSize);
		p_->packer.packInt(sendType);
		p_->packer.packInt(path.length()+1);
		p_->packer.pack(path.c_str(), path.length()+1);
		p_->packer.packInt(key.length()+1);
		p_->packer.pack(key.c_str(), key.length()+1);
		p_->packer.packInt(size);
		p_->packer.pack((char *)data, size);
		sendBuffer();
		}
	}


void ygNetClient::fetch(const ygString& path,const ygString& key)
	{
	if (p_->udpSocket)
		{
		int totalSize = sizeof(int) +
				sizeof(int) + path.length()+1 +
				sizeof(int) + key.length()+1;
		initPacker(totalSize);
		p_->packer.packInt(YG_NET_FETCH_KEY);
		p_->packer.packInt(path.length()+1);
		p_->packer.pack(path.c_str(), path.length()+1);
		p_->packer.packInt(key.length()+1);
		p_->packer.pack(key.c_str(), key.length()+1);
		sendBuffer();
		}
	}


void ygNetClient::process(void)
	{
	while (receivePacket())
		{
		int packetType = -1;
		p_->packer.unpackInt(&packetType);
		if ((packetType == YG_NET_PUT_KEY) ||
		    (packetType == YG_NET_REFRESH_KEY))
			{
			ygString path, key;
			if (processKeyData(path,key,packetType))
				{
				if (p_->callback)
					(*p_->callback)(path.c_str(),key.c_str(),
							p_->callbackData);
				}
			}
		else if (packetType == YG_NET_FETCH_KEY)
			processFetch();
		else
			; /* Ignore it */
//			cerr << "ERROR: ygNetClient::process received unknown"
//				" packet type " << packetType << endl;
		}
	refreshKeys();
	}


bool ygNetClient::receivePacket(void)
	{
	if (receiveTCPPacket())
		return true;
	return receiveUDPPacket();
	}


bool ygNetClient::processKeyData(ygString& path,ygString& key,int packetType)
	{
	int length;
	char *tmpstr;
/* Get path string length & data */
	p_->packer.unpackInt(&length);
	tmpstr = (char *) malloc(length);
	p_->packer.unpack(tmpstr, length);
	path.clear();
	path.append(tmpstr);
	free(tmpstr);
/* Get key string length & data */
	p_->packer.unpackInt(&length);
	tmpstr = (char *) malloc(length);
	p_->packer.unpack(tmpstr, length);
	key.clear();
	key.append(tmpstr);
	free(tmpstr);
/* Get data length & data */
	p_->packer.unpackInt(&length);
	ygNetClientKey * keystruct = getKeyStruct(path,key,false);
	if (keystruct)
		{
		bool returnValue = true;
		int oldchecksum, oldsize;
		p_->dataMutex.lock();
		if (packetType == YG_NET_REFRESH_KEY)
			{
			oldchecksum = computeChecksum(keystruct);
			oldsize = keystruct->size;
			}
		if (length > keystruct->bufferSize)
			{
			keystruct->bufferSize = length;
			keystruct->data = realloc(keystruct->data,keystruct->bufferSize);
			}
		p_->packer.unpack((char *)keystruct->data, length);
		keystruct->size = length;
		if (packetType == YG_NET_REFRESH_KEY)
			{
			if (length != oldsize)
				returnValue = false;
			else if (computeChecksum(keystruct) != oldchecksum)
				returnValue = false;
			}
		p_->dataMutex.unlock();
		return returnValue;
		}
	else
		return false;
	}


int ygNetClient::computeChecksum(ygNetClientKey * key) const
	{
	int i, sum = 0;
	for (i=0; i < key->size; i++)
		sum += ((char *)key->data)[i];
	return sum;
	}


void ygNetClient::processFetch(void)
	{
/* The reflector will currently handle fetching all the initial data, so this
   function is unneeded */
	}


void ygNetClient::downloadViaTCP(const ygString& server,int port)
	{
	cout << "ygNetClient::downloadViaTCP: connecting to " << server << ":" << port << endl;
	CAVERNnet_tcpReflectorClient_c *sock = new CAVERNnet_tcpReflectorClient_c;
	if (sock->connectToServer(server.c_str(),port) ==
			CAVERNnet_tcpReflectorClient_c::FAILED)
		{
		cerr << "ERROR: ygNetClient::downloadViaTCP: failed to connect"
			" to cache server " << server << ":" << port << endl;
		return;
		}
	int nbytes;
	initPacker(sizeof(int));
	p_->packer.packInt(YG_NET_REQUEST_ALL);
	nbytes = p_->packer.getBufferFilledSize();
	sock->write(p_->packer.getBuffer(),&nbytes);
	while (true)
		{
		char * buf;
		if (sock->read(&buf, &nbytes, CAVERNnet_tcpReflectorClient_c::BLOCKING)
				!= CAVERNnet_tcpReflectorClient_c::OK)
			break;
		p_->packer.initUnpack(buf,nbytes);
		int packetType = -1;
		p_->packer.unpackInt(&packetType);
		if (packetType == YG_NET_DOWNLOAD_DONE)
			{
			cout << "ygNetClient::downloadViaTCP: download done\n";
			delete buf;
			break;
			}
		else if (packetType == YG_NET_PUT_KEY)
			{
			ygString path, key;
			processKeyData(path,key);
			if (p_->debugDownload)
				cout << "ygNetClient::downloadViaTCP: received " << path << "/" << key << endl;
			}
		delete buf;
		}
	sock->close();
	}


void ygNetClient::uploadViaTCP(void)
	{
	cout << "ygNetClient::uploadViaTCP: sending all keys to "
		<< p_->server << ":" << p_->port << endl;
	ygNetClientPathmap::const_iterator iter;
	for (iter = p_->paths.begin(); iter != p_->paths.end(); ++iter)
		{
		ygNetClientKeymap::const_iterator iter2 = iter->second.begin();
		if (iter2->second->owned)
			{
			for ( ; iter2 != iter->second.end(); ++iter2)
				sendKeyDataTCP(iter->first, iter2->first,
					    iter2->second->data, iter2->second->size,
					    YG_NET_REFRESH_KEY);
			}
		}
	cout << "ygNetClient::uploadViaTCP: upload finished\n";
	}


void ygNetClient::refreshKeys(void)
	{
	if (p_->refreshInterval < 0)
		return;
	float t = ygGetTime();
	if ((t - p_->refreshTime) < p_->refreshInterval)
		return;
	p_->refreshTime = t;
/* Look for the next 'path' that we own locally (this presumes all keys within a particular
   path have the same 'owned' value, which is true as long as a path corresponds to an ygNode */
	while (p_->refreshIter != p_->paths.end())
		{
		ygNetClientKeymap::const_iterator iter2 = p_->refreshIter->second.begin();
/* When we find one, send out all its keys, then quit the outer while-loop */
		if (iter2->second->owned)
			{
			for ( ; iter2 != p_->refreshIter->second.end(); ++iter2)
				sendKeyDataUDP(p_->refreshIter->first, iter2->first,
					    iter2->second->data, iter2->second->size,
					    YG_NET_REFRESH_KEY);
			break;
			}
		else
			++(p_->refreshIter);
		}
	if (p_->refreshIter == p_->paths.end())
		p_->refreshIter = p_->paths.begin();
	else
		++(p_->refreshIter);
	}


void ygNetClient::initPacker(size_t s)
	{
	if (s > YG_NET_MAX_BUFFER_SIZE)
		{
		cerr << "ERROR: ygNetClient::initPacker(): trying to set buffer "
			"size to " << s << " bytes; maximum is " <<
			YG_NET_MAX_BUFFER_SIZE << endl;
		s = YG_NET_MAX_BUFFER_SIZE;
		}
	p_->packer.initPack((char*)p_->packetBuffer, s);	
	}


void ygNetClient::sendBuffer(void)
	{
	if (p_->udpSocket)
		{
		p_->udpSocket->send(p_->packer.getBuffer(),p_->packer.getBufferFilledSize());
		}
	}


bool ygNetClient::receiveTCPPacket(void)
	{
	if ((p_->tcpSocket) && (p_->tcpSocket->isReadyToRead() ==
				CAVERNnet_tcpReflectorClient_c::READY_TO_READ))
		{
		int numBytes=0;
		if (p_->tcpBuffer)
			delete p_->tcpBuffer;
		if (p_->tcpSocket->read(&p_->tcpBuffer, &numBytes,
					CAVERNnet_tcpReflectorClient_c::BLOCKING)
				== CAVERNnet_tcpReflectorClient_c::OK)
			{
			p_->packer.initUnpack(p_->tcpBuffer,numBytes);
			return true;
			}
		p_->tcpBuffer = NULL;
		}
	return false;
	}


bool ygNetClient::receiveUDPPacket(void)
	{
	if (p_->udpSocket)
		{
		int numBytes = p_->udpSocket->receive((char*)p_->packetBuffer,
							YG_NET_MAX_BUFFER_SIZE);
		if (numBytes > 0)
			{
			p_->packer.initUnpack((char *)p_->packetBuffer, numBytes);
			return true;
			}
		}
	return false;
	}
