// Description: creates a positional sound within a space using one of several file formats
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> this node can also be created with the alias: sound
//<li> the volume of a sound is the region in which the sound will play at 100% amplitude(loudness)
//<li> falloffDistance is a region outside the volume in which the amplitude is diminished as the user moves further away from the source(volume)
//<li> Common file formats supported by snerd:
//<b> .aiff</b> - Audio Interchange File Format
//<b> .wav</b> - MS RIFF WAVE Format
//<li> sound files should be saved using the <b>.aiff</b> format for best performance
//<ol> 
//<li> the sfconvert utility is located in /usr/sbin/
//<li> type sfconvert then the name of the file you want to convert, followed by the name of the converted file
//<li> <b>ex:</b> sfconvert oldfile.wav newfile.aif
//<li> see the man pages on sfconvert for more information
//</ol>
//<li> the sound server, snerd, must be run on the a local or remote machine
//<li> the location of sound files is set by the YG_SOUND_DIR environmental variable
//<li> sounds under a switch will turn off but must be restarted when the switch is toggled back on
//</ul>
//
// Category: Sound
// Author: Dave Pape
// Revision: 11/01/01
// Revision: 10/03/02 - Dave Pape - Added 'start' event.
//
#include <stdio.h>
#include <bergenSample.h>
#include <bergenServer.h>
#include <vector>
#include "ygVolume.h"
#include "ygUtil.h"
#include "ygWorld.h"
#include "ygNetKeys.h"
#include "ygSound.h"

using namespace std;

extern "C" ygNode* construct_ygSound(const char* name,bool master) { return new ygSound(name,master); }

struct _ygSoundPrivateData
	{
	ygString file;
	bergenSample * sample;
	float duration;
	float falloffDistance;
	float amplitude;
	bool loop;
	float startTime, endTime;
	float curAmplitude;
	bool isPlaying;
	unsigned int lastHeartbeat;
	ygString netFile;
	bool netPlay;
	};


struct _ygSoundGlobalData
	{
	bergenServer * server;
	vector<ygSound*> sounds;
	unsigned int currentHeartbeat;
	};

struct _ygSoundGlobalData * ygSound::global_ = NULL;



ygSound::ygSound(const char* name,bool master) : ygSpace(name,master)
	{
	setClassName("ygSound",true);
	if (!global_)
		initServer();
	p_ = new struct _ygSoundPrivateData;
	p_->sample = NULL;
	p_->falloffDistance = 10;
	p_->amplitude = 1;
	p_->loop = false;
	p_->startTime = 0;
	p_->endTime = 0;
	p_->curAmplitude = 0;
	p_->isPlaying = false;
	addSound(this);
	addNetKey("file", &p_->netFile, YG_NET_STRING);
	addNetKey("ampl", &p_->amplitude, YG_NET_FLOAT);
	addNetKey("dist", &p_->falloffDistance, YG_NET_FLOAT);
	addNetKey("loop", &p_->loop, YG_NET_BOOL);
	addNetKey("play", &p_->netPlay, YG_NET_BOOL);
	}

ygSound::~ygSound(void)
	{
	if (p_->sample)
		delete p_->sample;
	delete p_;
	}


void ygSound::reset(void)
	{
	//set fall off distance to 10.0
	p_->falloffDistance = 10;
	//set amplitude to 1.0
	p_->amplitude = 1;
	//reset loop flag to false
	p_->loop = false;
	//set start time to 0.0
	p_->startTime = 0;
	//set end time to 0.0
	p_->endTime = 0;
	//set current amplitude to 0.0
	p_->curAmplitude = 0;
	//reset playing flag to false
	p_->isPlaying = false;
	//if a sample exists then delete it
	if (p_->sample)
		{
		stop();
		delete p_->sample;
		p_->sample = NULL;
		}
	ygSpace::reset();
	}


void ygSound::message(const ygMessage& msg)
	{
	//set the name of the sound file
	if (msg == "file")
		{
		if (msg.args.size() > 0)
			setFile(msg.args[0]);
		}
	//set the zero amplitude distance from the space perimeter
	else if (msg == "falloffDistance")
		setFalloffDistance(msg.floatArg(0));
	//set the amplitude within the space
	else if (msg == "amplitude")
		setAmplitude(msg.floatArg(0));
	//set the sound looping mode
	else if (msg == "loop")
		setLoop(msg.boolArg(0));
	//start the sound playing
	else if (msg == "play")
		{
		if (msg.args.size() > 0)
			setFile(msg.args[0]);
		if (msg.args.size() > 1)
			setAmplitude(msg.floatArg(1));
		play();
		}
	//stop the playing sound
	else if (msg == "stop")
		stop();
	else
		ygSpace::message(msg);
	}


void ygSound::setFile(const ygString& val)
	{
	if (p_->sample)
		{
		stop();
		delete p_->sample;
		p_->sample = NULL;
		}
	p_->file = val;
	p_->netFile = p_->file;
	netKeyChanged("file");
	}


void ygSound::setFalloffDistance(float val)
	{
	p_->falloffDistance = val;
	netKeyChanged("dist");
	}


void ygSound::setAmplitude(float val)
	{
	p_->amplitude = val;
	netKeyChanged("ampl");
	}


void ygSound::setLoop(bool val)
	{
	p_->loop = val;
	if (p_->sample)
		p_->sample->setLoop(val);
	netKeyChanged("loop");
	}


void ygSound::setCurAmplitude(float val)
	{
	p_->curAmplitude = val;
	if (p_->sample)
		p_->sample->setAmplitude(val);
	}


bool ygSound::isPlaying(void)
	{
	return p_->isPlaying;
	}


void ygSound::play(void)
	{
	if (p_->isPlaying)
		return;
	//if no sample then create a new bergenSample
	if (!p_->sample)
		{
		p_->sample = new bergenSample(p_->file.c_str(),server());
		p_->duration = p_->sample->duration();
		if (p_->loop)
			p_->sample->setLoop(1);
		else
			p_->sample->setLoop(0);
		}
	p_->sample->setLoop(p_->loop);
	setCurAmplitude(computeAmplitude());
	//play the sample
	p_->sample->play();
	//generate a heartbeat for the server
	heartbeat();
	p_->isPlaying = true;
	if (!p_->loop)
		{
		p_->startTime = ygWorld::FrameTime;
		p_->endTime = p_->startTime + p_->duration;
		}
	eventOccurred("start");
	p_->netPlay = true;
	netKeyChanged("play");
	}


void ygSound::stop(bool sendStopCommand)
	{
	//if a sample is playing
	if (p_->isPlaying)
		{
		//send a stop command to the server
		if ((sendStopCommand) && (p_->sample))
			{
			p_->sample->stop();
			setCurAmplitude(0.0);
			}
		p_->isPlaying = false;
		eventOccurred("stop");
		p_->netPlay = false;
		netKeyChanged("play");
		}
	}


void ygSound::heartbeat(void)
	{
	if (global_)
		p_->lastHeartbeat = global_->currentHeartbeat;
	}


void ygSound::app(void)
	{
	//if a sample exists
	if (p_->sample)
		{
		//generate a heartbeat for the server
		heartbeat();
		//if the sample is playing
		if (p_->isPlaying)
			{
			//if sample is past end time and not looping then stop
			if ((!p_->loop) && (p_->endTime < ygWorld::FrameTime))
				stop(false);
			//else, compute new amplitude
			else
				{
				float newAmpl = computeAmplitude();
				if (fabs(newAmpl - p_->curAmplitude) > 0.005)
					setCurAmplitude(newAmpl);
				}
			}
		}
	ygSpace::app();
	}


float ygSound::computeAmplitude(void)
	{
	float amp = p_->amplitude;
	float maxDist = p_->falloffDistance;
	ygNode * viewer = ygWorld::World->viewer();
	if ((maxDist >= 0) && (viewer))
		{
		viewer->app();
		pfVec3 userPos = viewer->origin(this);
		float distSq;
		if (volume())
			distSq = volume()->sqrDistance(userPos);
		else
			distSq = userPos.dot(userPos);
		if (distSq > maxDist*maxDist)
			amp = 0;
		else if (maxDist > 0)
			amp *= (1.0f - sqrtf(distSq) / maxDist);
		}
	return amp;
	}


void ygSound::acceptNetKey(const ygString& key)
	{
	if (key == "file")
		setFile(p_->netFile);
	else if (key == "play")
		{
		if (p_->netPlay)
			play();
		else
			stop();
		}
	else
		ygSpace::acceptNetKey(key);
	}


/****** Global functions for connection to server & heartbeat *******/

void ygSound::initServer(void)
	{
	if (!global_)
		{
		global_ = new struct _ygSoundGlobalData;
		global_->server = new bergenServer;
		if (getenv("YG_SOUND_DIR"))
			global_->server->setDirectory(getenv("YG_SOUND_DIR"));
		global_->currentHeartbeat = 0;
		}
	}


void ygSound::addSound(ygSound* s)
	{
	if (global_)
		global_->sounds.push_back(s);
	}


void ygSound::removeSound(ygSound* s)
	{
	if (global_)
		{
		vector<ygSound*>::iterator iter;
		for (iter = global_->sounds.begin(); iter != global_->sounds.end();
		    ++iter)
			if (*iter == s)
				{
				global_->sounds.erase(iter);
				break;
				}
		}
	}


void ygSound::updateServer(void)
	{
	vector<ygSound*>::iterator iter;
	for (iter = global_->sounds.begin(); iter != global_->sounds.end(); ++iter)
		if (((*iter)->p_->isPlaying) &&
		    ((*iter)->p_->lastHeartbeat < global_->currentHeartbeat))
			(*iter)->stop();
	global_->currentHeartbeat++;
	}


void ygSound::closeServer(void)
	{
	delete global_->server;
	delete global_;
	global_ = NULL;
	}


bergenServer * ygSound::server(void)
	{
	return global_->server;
	}
